#! /usr/bin/python
# -*- coding: UTF-8 -*-
"""Indicator-Bug lists bugs from a Launchpad project in an indicator and and updates it regularly.
It supports Launchpads' usual advanced bug searching options.
Clicking on a bug opens the bugs page on a browser"""

# Copyright (c) 2011 Oier Mees
#
# indicator-bug is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any
# later version.
#
#
# You should have received a copy of the GNU General Public License
# along with indicator-bug; see the file COPYING.  If not, write to the Free
# Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
# Authors:
#  Oier Mees <oier.mees@gmail.com>

# std
import os
import sys
import gettext
import locale
import webbrowser
import logging, logging.handlers
import commands
import shutil
import tempfile

# local
from indicatorBug import preferences, prompts

# 3rd party
from gi.repository import Gtk, GObject
from gi.repository import AppIndicator3 as appindicator
from launchpadlib.launchpad import Launchpad
from launchpadlib.errors import HTTPError

PING_FREQUENCY = 300
# Set up translations
locale.setlocale(locale.LC_ALL, '')
_ = gettext.gettext
gettext.textdomain('indicator-bug')

# Add project root directory (enable symlink, and trunk execution).
PROJECT_ROOT_DIRECTORY = os.path.abspath(
    os.path.dirname(os.path.dirname(os.path.realpath(sys.argv[0]))))

VERSION = '0.2'


class indicatorBug:

        def __init__(self):
            self.initializeLogger()
            self.ind = appindicator.Indicator.new("indicator-bug",
                                               "indicator-bug-status",
                                               appindicator.IndicatorCategory.APPLICATION_STATUS)
            self.ind.set_status(appindicator.IndicatorStatus.ACTIVE)
            self.ind.set_attention_icon("indicator-bug-new")
            self.login()
            self.prefs = preferences.PreferencesDialog(self.launchpad, log)
            self.oldBugList = []
            self.check_bugs()
            self.menu_setup()
            self.ind.set_menu(self.menu)
            self.prefs.connect("data-change", self.update_menu)

        def initializeLogger(self):
            global log
            log_filename = os.path.join(os.path.expanduser("~/.cache"), "indicator-bug.log")
            log = logging.getLogger('indicator-bug')
            log.propagate = False
            log.setLevel(logging.DEBUG)
            log_handler = logging.handlers.RotatingFileHandler(log_filename, maxBytes=1024*1024, backupCount=5)
            log_formatter = logging.Formatter("[%(threadName)s] %(asctime)s - %(levelname)s - %(message)s")
            log_handler.setFormatter(log_formatter)
            log.addHandler(log_handler)

            log.info("------------------------------")
            log.info("Started indicator-bug from %s" % PROJECT_ROOT_DIRECTORY)
            log.info("Indicator-bug version %s" % VERSION)
            self.is_single_instance()

        def is_single_instance(self):
            # Single instance stuff for indicator-bug
            self.myapp = SingleInstance("/tmp/indicator-bug-%d.pid" % os.getuid())
            # check is another instance of same program running
            if self.myapp.is_already_running():
                message=_("Another instance of this program is already running.")
                prompts.error(_("Error occured!"), message)
                log.info(message)
                sys.exit(message)

        def update_timer(self):
            GObject.timeout_add(PING_FREQUENCY * 1000, self.update_timer)
            self.update_menu(None, None)

        def update_menu(self, object, var):
            self.check_bugs()
            self.menu_setup()
            self.ind.set_menu(self.menu)

        def add(self, text = None, icon = None, conector_event = None, conector_action = None, cb_args=None):
            if text and icon is None:
                menu_item = Gtk.MenuItem(text)
                log.debug("adding \"%s\" to the indicator"% text)
            elif text and icon:
                menu_item = Gtk.ImageMenuItem(text)
                image = Gtk.Image()
                image.set_from_file(icon)
                menu_item.set_image(image)
                menu_item.set_always_show_image(True)
                log.debug("adding \"%s\" to the indicator"% text)
            else:
                menu_item = Gtk.SeparatorMenuItem()
                log.debug("adding SeparatorMenuItem to the indicator")
            if conector_event and conector_action:
                if (cb_args):
                    menu_item.connect(conector_event, conector_action, cb_args)
                else:
                    menu_item.connect(conector_event, conector_action)
            menu_item.show()
            self.menu.append(menu_item)
            return menu_item

        def menu_setup(self):
            log.debug("setting up the menu...")
            self.menu = Gtk.Menu()
            self.add(self.project.title()+" Bugs", conector_event = 'activate', conector_action = self.openBugWeb, cb_args=self.prefs.project)
            self.add()
            if(len(self.bugList)==0):
                self.add(text=_("No bugs found!"))
                self.add(text=_('Please try changing the preferences'), conector_event = 'activate', conector_action = self.menu_set_preferences)
                log.debug("No bugs found!")
            i = 0
            for bug in self.bugList[:8]:
                title = ((self.bugList[i].title).partition(':')[2]).strip('\" ')
                if(len(title)>45):
                    title = title[:45]+"..."
                self.add(title, conector_event = 'activate', conector_action = self.openBugWeb, cb_args=bug)
                i = i+1
                title = ""
            self.add()
            menu_preferences = self.add(text = _('Preferences'), conector_event = 'activate', conector_action = self.menu_set_preferences)
            self.quit_item = Gtk.MenuItem(_("Quit"))
            self.quit_item.connect("activate", self.quit)
            self.quit_item.show()
            self.menu.append(self.quit_item)
            self.menu.show_all()

        def openBugWeb(self, data, bug):
            log.debug("opening the web browser...")
            webbrowser.open(bug.web_link)

        def menu_set_preferences(self, widget):
            return self.prefs.run()

        def main(self):
            GObject.timeout_add(PING_FREQUENCY * 1000, self.update_timer)
            Gtk.main()

        def quit(self, widget):
            log.debug("exiting...")
            sys.exit(0)

        def login(self):
            log.debug("trying to establish connection with Launchpad...")
            try:
                self.launchpad = Launchpad.login_anonymously("indicator-bug", "production")
            except Exception, error:
                prompts.error(_("Error occured!"), _("Could not connect to Launchpad."))
                log.error("Could not connect to Launchpad")
                sys.exit(2)

        def load_preferences(self):
            log.debug("loading preferences...")
            self.project = self.prefs.settings['project_name']
            self.tagsList = self.prefs.settings['tags']
            if not self.tagsList:
                self.tagsList = "*"
            self.checkbuttonUndecided = self.prefs.settings['checkbuttonUndecided']
            self.checkbuttonMedium = self.prefs.settings['checkbuttonMedium']
            self.checkbuttonWishlist = self.prefs.settings['checkbuttonWishlist']
            self.checkbuttonHigh = self.prefs.settings['checkbuttonHigh']
            self.checkbuttonLow = self.prefs.settings['checkbuttonLow']
            self.checkbuttonCritical = self.prefs.settings['checkbuttonCritical']
            self.checkbuttonNew = self.prefs.settings['checkbuttonNew']
            self.checkbuttonIncomplete = self.prefs.settings['checkbuttonIncomplete']
            self.checkbuttonTriaged = self.prefs.settings['checkbuttonTriaged']
            self.checkbuttonInProgress = self.prefs.settings['checkbuttonInProgress']
            self.checkbuttonFixCommited = self.prefs.settings['checkbuttonFixCommited']
            self.checkbuttonConfirmed = self.prefs.settings['checkbuttonConfirmed']
            self.importanceList = []
            if self.checkbuttonUndecided:
                self.importanceList.append('Undecided')
            if self.checkbuttonMedium:
                self.importanceList.append('Medium')
            if self.checkbuttonWishlist:
                self.importanceList.append('Wishlist')
            if self.checkbuttonHigh:
                self.importanceList.append('High')
            if self.checkbuttonLow:
                self.importanceList.append('Low')
            if self.checkbuttonCritical:
                self.importanceList.append('Critical')

            self.statusList = []
            if self.checkbuttonNew:
                self.statusList.append('New')
            if self.checkbuttonIncomplete:
                self.statusList.append('Incomplete (with response)')
            if self.checkbuttonTriaged:
                self.statusList.append('Triaged')
            if self.checkbuttonInProgress:
                self.statusList.append('In Progress')
            if self.checkbuttonFixCommited:
                self.statusList.append('Fix Committed')
            if self.checkbuttonConfirmed:
                self.statusList.append('Confirmed')

        def check_bugs(self):
            self.load_preferences()
            projectResult = self.launchpad.projects[self.project]
            log.debug("Project: %s Tags:%s Importance:%s Status:%s" % (self.project,
            self.tagsList, self.importanceList, self.statusList))
            try:
                self.bugList = projectResult.searchTasks(tags=self.tagsList,
                importance=self.importanceList, status=self.statusList,
                omit_duplicates=True, tags_combinator='Any')
            except HTTPError, e:
                log.error("Error while fetching bugs: %s" % e.content)
            if((self.oldBugList and self.bugList) and
                (self.oldBugList[:8]!= self.bugList[:8])):
                    self.ind.set_status(appindicator.IndicatorStatus.ATTENTION)
                    log.debug("Setting indicator status to attention")
            else:
                self.ind.set_status(appindicator.IndicatorStatus.ACTIVE)
            self.oldBugList = self.bugList
            return True


class SingleInstance(object):
    """ Class to ensure, that single instance of the applet is run for each user """

    # Initialize, specifying a path to store pids
    def __init__(self, pidPath):
        self.pidPath=pidPath
        # See if pidFile exists
        if os.path.exists(pidPath):
            log.debug("SingleInstance: pid file %s exists" % pidPath)
            # Make sure it is not a "stale" pidFile
            pid=open(pidPath, 'r').read().strip()
            # Check list of running pids, if not running it is stale so overwrite
            pidRunning = commands.getoutput('ls /proc | grep %s' % pid)
            log.debug("SingleInstance: pid running %s" % pidRunning)
            self.lasterror = True if pidRunning else False
        else:
            self.lasterror = False

        if not self.lasterror:
            log.debug("SingleInstance: writing new pid %s" % str(os.getpid()))
            # Create a temp file, copy it to pidPath and remove temporary file
            (fp, temp_path)=tempfile.mkstemp()
            try:
                os.fdopen(fp, "w+b").write(str(os.getpid()))
                shutil.copy(temp_path, pidPath)
                os.unlink(temp_path)
            except Exception as e:
                log.error("SingleInstance: exception while renaming '%s' to '%s':\n %s" % (temp_path, pidPath, str(e)))

    def is_already_running(self):
        return self.lasterror

    def __del__(self):
        if not self.lasterror:
            log.debug("SingleInstance: deleting %s" % self.pidPath)
            os.unlink(self.pidPath)
